#pragma once

#include "weakptr.h"

class Room;
class EntityFactory;
class Explosion : public std::enable_shared_from_this<Explosion>
{
 public:
    ~Explosion();
    int exclude_uniid = 0;

    Room* GetRoom() { return room_; };
    CreatureWeakPtr GetSender() { return sender_; };
    float GetDmg() { return dmg_; };
    long long GetSpecialDamageType() { return special_damage_type_; };
    bool IsPreBattleExplosion();
    int GetExplosionEffect() { return explosion_effect_; };
    void SetDamageDelay(int delay) { explosion_damage_delay_ = delay; };

    void IndifferenceAttack(Room* room,
                            const a8::Vec2& center,
                            float explosion_range,
                            int explosion_effect,
                            float dmg,
                            long long special_damage_type = 0);
    void EnemyAndObstacleAttack(CreatureWeakPtr& sender,
                                const a8::Vec2& center,
                                float explosion_range,
                                int explosion_effect,
                                float dmg,
                                long long special_damage_type = 0);
    void AddForceTarget(CreatureWeakPtr force_target);

protected:
    Explosion() {};
    void InternalAttack();
    void ProcDamage();

 private:
    int type_ = 0;
    Room* room_ = nullptr;
    CreatureWeakPtr sender_;
    CreatureWeakPtr force_target_;
    float explosion_range_ = 0;
    int explosion_effect_ = 0;
    int explosion_damage_delay_ = 0;
    float dmg_ = 0;
    a8::Vec2 center_;
    long long special_damage_type_ = 0;
    long long create_frameno_ = 0;

    friend class EntityFactory;
};
